# TODO
class I3Binder:
    """Read key bindings and render nice info file to look up the bindings"""

    def fetch_bindings(self):
        """read in all bindcode and bindsym assignments"""

    def translate_bindings(self):
        """translate bindsym to bindsym assignments
    
        this need to be done the moment the information is asked because it
        depends on the currently acitve layout.
        """
    def write_bindings_info(self):
        pass



import logging
import pprint
import re
import time
from pathlib import Path
from string import Template

from i3configger import defaults

log = logging.getLogger(__name__)


class I3Configger:
    def __init__(self,
                 sourcePath=defaults.SOURCES_PATH,
                 targetPath=defaults.TARGET_PATH,
                 suffix=defaults.SOURCE_SUFFIX,
                 selectors=None, statusMarker=None):
        self.sourcePath = sourcePath
        self.mainTargetPath = targetPath
        self.suffix = suffix
        self.selectors = selectors or []
        self.statusMarker = statusMarker
        log.info("initialized %s", self)

    def __str__(self):
        return "%s(%s)" % (
            self.__class__.__name__, pprint.pformat(self.__dict__))

    def build(self):
        self.build_main_config()
        self.build_i3status()

    def build_main_config(self):
        """two pass rendering to figure out from content what"""
        paths = self.get_file_paths(
            self.sourcePath,
            selectors=self.selectors + [[self.statusMarker, None]])
        self.content = self.get_content(paths)
        self.vars = self.render_vars(self.content)
        config = self.render_config(self.content, self.vars)
        self.mainTargetPath.write_text(config)

    def build_i3status(self):
        settings = self.fetch_settings('i3status', self.vars)
        for _, target in settings.items():
            targetPath = Path(target).expanduser()
            srcPath = self.sourcePath / targetPath.name
            content = self.get_content(srcPath)
            status = self.render_config(content, self.vars)
            targetPath.write_text(status)

    @classmethod
    def render_config(cls, content, vars_):
        """Works out of the box, because $ is the standard substitution marker 
        for string.Template"""
        tpl = Template(content)
        renderedContent = tpl.safe_substitute(vars_)
        return renderedContent

    @classmethod
    def render_vars(cls, content):
        """resolve and remove $prefix for string substitution"""
        vars_ = cls.parse(content)
        resolvedVars = cls.resolve(vars_)
        return {key[1:]: value for key, value in resolvedVars.items()}

    @staticmethod
    def fetch_settings(marker, vars_):
        """Simple way of communicating settings to i3configger

        $i3configger_somecommonsetting_keyname value
        $i3configger_somecommonsetting_otherkeyname othervalue

        will be turned into ...
        
        dict(keyname=value, otherkeyname=othervalue)
        """
        marker = 'i3configger_' + marker + '_'
        markerLen = len(marker)
        return {k[markerLen:]: v for k, v in vars_.items()
                if k.startswith(marker)}

    def get_file_paths(self, sourcePath, selectors=None):
        """:returns: list of pathlib.Path"""
        filePaths = []
        for sp in [p for p in self.sourcePath.iterdir()]:
            if not sp.is_file():
                continue
            if sp.suffix != self.suffix:
                continue
            parts = sp.stem.split('.')
            if len(parts) > 1 and selectors:
                criterion, spec, *_ = parts
                for wanted, value in selectors:
                    if criterion == wanted:
                        if not value:
                            continue
                        if ((isinstance(spec, str) and spec == value)
                                or spec in value):
                            filePaths.append(sp)
            else:
                filePaths.append(sp)
        return sorted(filePaths)

    @staticmethod
    def resolve(vars_):
        resolvedVars = vars_.copy()
        """resolve values that are vars_ (e.g. set $var $ otherVar)"""
        for key, value in resolvedVars.items():
            if value.startswith('$'):
                try:
                    resolvedVars[key] = resolvedVars[value]
                except KeyError:
                    log.exception("[IGNORED] %s, %s", key, value)
        return resolvedVars

    @staticmethod
    def get_content(pathOrPaths):
        paths = [pathOrPaths] if isinstance(pathOrPaths, Path) else pathOrPaths
        msg = '# generated by i3configger (%s) #' % (time.asctime())
        sep = "#" * len(msg)
        out = ["%s\n%s\n%s\n" % (sep, msg, sep)]
        for filePath in paths:
            content = filePath.read_text()
            content = "### %s ###\n%s\n" % (filePath.name, content)
            out.append(content)
        return ''.join(out)

    @classmethod
    def parse(cls, content):
        """read all set commands"""
        vars_ = {}
        for line in [l.strip() for l in content.splitlines()]:
            if cls.looks_like_assignment(line):
                key, value = cls.get_assignment(line)
                vars_[key] = value
        return vars_

    @staticmethod
    def looks_like_assignment(line):
        return line.startswith('set ')

    ASSIGNMENT_RE = re.compile(r'(\$\S+)\s+(.*)')

    @classmethod
    def get_assignment(cls, line):
        sanitizedLine = (line.split('set')[1].split(' # ')[0]).strip()
        if len([c for c in sanitizedLine if c == '#']) > 1:
            raise MalformedAssignment("comments need space: '%s'", line)
        match = re.match(cls.ASSIGNMENT_RE, sanitizedLine)
        if not match or len(match.groups()) != 2:
            raise MalformedAssignment("can't match properly: '%s'", line)
        return match.group(1), match.group(2)


class MalformedAssignment(Exception):
    pass



import logging
import os
import subprocess

log = logging.getLogger(__name__)
DEBUG = os.getenv('DEBUG', 0)


def dbg_print(msg, *args):
    if not DEBUG:
        return
    if args:
        msg %= args
    print(msg)


def get_selectors(parser, argv):
    selectors = []
    unrecognized = []
    marker = '--select-'
    markerLen = len(marker)
    for arg in argv:
        if not arg.startswith(marker):
            unrecognized.append(arg)
        if '=' not in arg:
            parser.error("arg '%s': need format --select-key=value" % (arg))
        selectors.append(arg[markerLen:].split('='))
    if unrecognized:
        parser.error('unrecognized arguments: %s' % ' '.join(argv))
    return selectors


def configure_logging(verbosity, logpath, isDaemon=False):
    level = logging.getLevelName(
        {0: 'ERROR', 1: 'WARNING', 2: 'INFO'}.get(verbosity, 'DEBUG'))
    fmt = '%(asctime)s %(name)s %(levelname)s: %(message)s'
    if isDaemon:
        logging.basicConfig(filename=str(logpath), format=fmt, level=level)
    else:
        logging.basicConfig(format=fmt, level=level)
        fileHandler = logging.FileHandler(str(logpath))
        fileHandler.setFormatter(logging.Formatter(fmt))
        fileHandler.setLevel(level)
        log.addHandler(fileHandler)


class IpcControl:
    @classmethod
    def reload_i3(cls):
        if cls._send_i3_msg('reload'):
            cls.notify_send("reloaded i3")

    @classmethod
    def restart_i3(cls):
        if cls._send_i3_msg('restart'):
            cls.notify_send("restarted i3")

    refresh = restart_i3

    @classmethod
    def _send_i3_msg(cls, msg):
        # todo use Adaephons i3 library
        cmd = ['i3-msg', msg]
        try:
            output = subprocess.check_output(cmd).decode()
            if '"success":true' in output:
                return True
            cls.notify_send("%s: %s" % (cmd, output), urgency='critical')
            return False
        except subprocess.CalledProcessError as e:
            if msg == 'restart' and e.returncode == 1:
                log.debug("[IGNORE] exit 1 is ok for restart")
                return True

    @classmethod
    def notify_send(cls, msg, urgency='low'):
        subprocess.check_call([
            'notify-send', '-a', 'i3configger', '-t', '1', '-u', urgency, msg])



import argparse
import logging
import sys

from i3configger import __version__, daemonize, defaults, util
from i3configger.configger import I3Configger
from i3configger.watch import Watchman

log = logging.getLogger(__name__)


def parse_args():
    p = argparse.ArgumentParser(
        'i3configger', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('--version', action='version', version=__version__)
    p.add_argument('-v', action="count", help="raise verbosity", default=0)
    p.add_argument('--sources', action="store",
                   default=defaults.SOURCES_PATH, help="path to sources")
    p.add_argument('--target', action="store",
                   default=defaults.TARGET_PATH,
                   help="path to main config file")
    p.add_argument('--suffix', action="store",
                   default=defaults.SOURCE_SUFFIX,
                   help="suffix of config source files")
    p.add_argument('--reload', action="store_true", default=False,
                   help="reload i3 instead of restart (not i3bar update)")
    p.add_argument('--logpath', action="store", default=defaults.LOG_PATH,
                   help="log to given path")
    g = p.add_mutually_exclusive_group()
    g.add_argument('--watch', action="store_true",
                   help="watch and build in foreground", default=False)
    g.add_argument('--daemon', action="store_true",
                   help="watch and build as daemon", default=False)
    g.add_argument('--kill', action="store_true", default=False,
                   help="exorcise daemon if running")

    # fixme there has to be a better way - integrate into selectors
    p.add_argument('--statusmarker', action="store",
                   default=defaults.STATUS_MARKER,
                   help="prefix of status bar configurations")

    args, argv = p.parse_known_args()
    args.selectors = util.get_selectors(p, argv)
    util.dbg_print("initialized with %s", args)
    return args


def main():
    args = parse_args()
    configgerArgs = (args.sources, args.target, args.suffix, args.selectors,
                     args.statusmarker)
    if args.daemon:
        daemonize.daemonize(args.v, args.logpath, configgerArgs)
    if args.kill:
        daemonize.exorcise()
        return 0
    if args.reload:
        util.IpcControl.refresh = util.IpcControl.reload_i3
    util.configure_logging(verbosity=args.v, logpath=args.logpath)
    if args.watch:
        try:
            Watchman(configgerArgs).watch()
        except KeyboardInterrupt:
            sys.exit("bye")
    else:
        configger = I3Configger(*configgerArgs)
        configger.build()
        # TODO need a way to refresh i3bar config without restarting i3
        util.IpcControl.refresh()


if __name__ == '__main__':
    sys.exit(main())



import os
import sys
from pathlib import Path

import daemon
import psutil

from i3configger.watch import Watchman
from i3configger.util import configure_logging


def get_other_i3configgers():
    """should always be max one, but you never know ..."""
    others = [p for p in psutil.process_iter() if p.name() == 'i3configger']
    return [p for p in others if p.pid != os.getpid()]


def daemonize(verbosity, logpath, configgerArgs):
    others = get_other_i3configgers()
    if others:
        sys.exit("i3configger already running (%s)" % others)
    context = daemon.DaemonContext(
        working_directory=Path(__file__).parent
        # TODO check if this umask is ok umask=0o002
    )
    # todo handle signals properly
    # context.signal_map = {
    #     signal.SIGTERM: program_cleanup,
    #     signal.SIGHUP: 'terminate',
    #     signal.SIGUSR1: reload_program_config}
    if verbosity:
        # spew output to terminal from where daemon was started
        context.stdout = sys.stdout
        context.stderr = sys.stderr
    with context:
        configure_logging(verbosity, logpath, isDaemon=True)
        Watchman(configgerArgs).watch_guarded()


def exorcise():
    # todo some error handling
    for process in get_other_i3configgers():
        print("killing %s" % process.pid)
        process.kill()



import tempfile
from pathlib import Path

LOG_PATH = Path(tempfile.gettempdir()) / 'i3configger.log'
SOURCES_PATH = Path('~/.i3/config.d').expanduser()
TARGET_PATH = Path('~/.i3/config').expanduser()
SOURCE_SUFFIX = '.conf'
STATUS_MARKER = 'i3status'
